@define-color black  #000000;
@define-color gray18 #2e2e2e;
@define-color gray22 #383838;
@define-color gray26 #424242;
@define-color gray32 #525252;
@define-color gray40 #666666;
@define-color gray46 #757575;
@define-color white  #ffffff;

.progress-overlay
{
    min-height: 1px;
}

.progress-overlay trough
{
    border-radius: 0px;
    border-color: transparent;
    border-width: 1px 0px;
    min-height: 3px;
}

.progress-overlay progress
{
    border-radius: 0px;
    border-width: 1px;
    min-height: 3px;
}

.ghb-preview-hud
{
    border-radius: 12px;
    border-width: 1px;
    padding: 6px;
}

.ghb-preview-image-frame
{
    background-color: black;
}

.ghb-button > box
{
    padding-left: 4px;
    padding-right: 4px;
}

.ghb-button > box.horizontal > label
{
    margin: 0px;
    padding-right: 0px;
    padding-left: 6px;
}

.ghb-button > box.vertical > label
{
    margin: 0px;
    padding: 0px;
}

.ghb-button > box image
{
    margin: 0px;
}

.ghb-button .ghb-indicator
{
    background-color: #3584e4;
    color: white;
    border-radius: 100px;
    padding: 0px 4px;
    margin: -6px -7px;
    font-size: 7pt;
    font-weight: 800;
}

.large-icons > .ghb-button .ghb-indicator
{
    padding: 1px 6px;
    margin: -4px -6px;
    font-size: 8pt;
}

.ghb-window-title
{
    margin: -4px 4px;
}

.numeric
{
    font-feature-settings: "tnum";
}

.subtitle
{
    font-size: smaller;
    opacity: 0.6;
}

textview
{
    padding: 5px 5px 5px 5px;
}

.entry
{
    margin: 0px 5px 0px 5px;
    padding: 0px 0px 0px 0px;
}

stackswitcher button.text-button
{
    min-width: 50px;
}

.ghb-monospace
{
    font-size: 8pt;
    padding: 0px 5px;
}

.row.drag-hover-top
{
    box-shadow: 0px 1px #4e9a06 inset;
}

.row.drag-hover-top:first-child
{
    box-shadow: 0px 2px #4e9a06 inset;
}

.row.drag-hover-bottom
{
    box-shadow: 0px -1px #4e9a06 inset;
}

.row.drag-hover-bottom:last-child
{
    box-shadow: 0px -2px #4e9a06 inset;
}

toolbar
{
    -gtk-icon-style: symbolic;
}

.ghb-queue-panel
{
    min-width: 333pt;
}

.ghb-linked-button > button:dir(ltr):not(:first-child),
.ghb-linked-button > button:dir(rtl):not(:last-child),
.ghb-linked-button > menubutton:dir(ltr):not(:first-child) > button,
.ghb-linked-button > menubutton:dir(rtl):not(:last-child) > button
{
    border-top-left-radius: 0;
    border-bottom-left-radius: 0;
    margin-left: -1px;
}

.ghb-linked-button > button:dir(ltr):not(:last-child),
.ghb-linked-button > button:dir(rtl):not(:first-child),
.ghb-linked-button > menubutton:dir(ltr):not(:last-child) > button,
.ghb-linked-button > menubutton:dir(rtl):not(:first-child) > button
{
    border-top-right-radius: 0;
    border-bottom-right-radius: 0;
}

.ghb-linked-button > menubutton > button
{
    padding-left: 1px;
    padding-right: 1px;
}

.ghb-preview-window
{
    border-bottom-left-radius: 0;
    border-bottom-right-radius: 0;
}
