const char hb_builtin_presets_json[] =
"{\n"
"    \"PresetBuiltin\": [\n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"fast\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureLooseCrop\": false,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 1,\n"
"                    \"PicturePARWidth\": 1,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"AV1 video (up to 2160p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Very Fast 2160p60 4K AV1\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 7500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"svt_av1_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"enable-variance-boost=1:enable-qm=1:ac-bias=1:tf-strength=1:qp-scale-compress-strength=1:sharpness=1:keyint=10s\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"7\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 43.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"psnr\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"fast\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureLooseCrop\": false,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 1,\n"
"                    \"PicturePARWidth\": 1,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Very Fast 2160p60 4K HEVC\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 7500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"superfast\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 26.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"fast\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"Small H.264 video (up to 1080p30) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Very Fast 1080p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 4000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"veryfast\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 24.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"fast\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"Small H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Very Fast 720p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 2000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"veryfast\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 23.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"fast\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 576,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"Small H.264 video (up to 576p25) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Very Fast 576p25\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1200,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"25\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"veryfast\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"fast\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 480,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"Small H.264 video (up to 480p30) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Very Fast 480p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"veryfast\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureLooseCrop\": false,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 1,\n"
"                    \"PicturePARWidth\": 1,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"AV1 video (up to 2160p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Fast 2160p60 4K AV1\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 10500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"svt_av1_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"enable-variance-boost=1:enable-qm=1:ac-bias=1:tf-strength=1:qp-scale-compress-strength=1:sharpness=1:keyint=10s\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"6\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 34.5,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"psnr\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureLooseCrop\": false,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 1,\n"
"                    \"PicturePARWidth\": 1,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Fast 2160p60 4K HEVC\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 13000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"faster\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 24.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": true,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Fast 1080p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"fast\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Fast 720p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 3000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"fast\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 21.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 576,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"H.264 video (up to 576p25) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Fast 576p25\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1800,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"25\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"fast\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 480,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"H.264 video (up to 480p30) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Fast 480p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"fast\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureLooseCrop\": false,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 1,\n"
"                    \"PicturePARWidth\": 1,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"High quality AV1 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"HQ 2160p60 4K AV1 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 16000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"svt_av1_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"enable-variance-boost=1:enable-qm=1:ac-bias=1:tf-strength=1:qp-scale-compress-strength=1:sharpness=1:keyint=10s\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"4\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 30.5,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"psnr\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureLooseCrop\": false,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 1,\n"
"                    \"PicturePARWidth\": 1,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"High quality H.265 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"HQ 2160p60 4K HEVC Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 20500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"HQ 1080p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 9000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 720p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"HQ 720p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 4500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 19.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 576,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 576p25), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"HQ 576p25 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 2700,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"25\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 18.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 480,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"High quality H.264 video (up to 480p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"HQ 480p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 2250,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 18.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureLooseCrop\": false,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 1,\n"
"                    \"PicturePARWidth\": 1,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"Super high quality AV1 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Super HQ 2160p60 4K AV1 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 32000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"svt_av1_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"enable-variance-boost=1:enable-qm=1:ac-bias=1:tf-strength=1:qp-scale-compress-strength=1:sharpness=1:keyint=10s\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"3\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 25.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"psnr\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureLooseCrop\": false,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 1,\n"
"                    \"PicturePARWidth\": 1,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"Super high quality H.265 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Super HQ 2160p60 4K HEVC Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 32000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"Super high quality H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Super HQ 1080p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 15000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"ref=5:bframes=5\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"veryslow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 18.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"Super high quality H.264 video (up to 720p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Super HQ 720p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 7500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"ref=5:bframes=5\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"veryslow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 17.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 576,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"Super high quality H.264 video (up to 576p25), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Super HQ 576p25 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 4500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"25\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"ref=5:bframes=5\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"veryslow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 16.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 480,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"Super high quality H.264 video (up to 480p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Super HQ 480p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 3750,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"ref=5:bframes=5\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"veryslow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 16.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ],\n"
"            \"Folder\": true,\n"
"            \"PresetName\": \"General\",\n"
"            \"Type\": 0\n"
"        },\n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"48\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"High quality video for publishing via online services such as Vimeo and YouTube. H.264 video (up to 2160p60) and high bit rate AAC stereo audio in an MP4 container.\",\n"
"                    \"PresetName\": \"Creator 2160p60 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 18000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"5.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"keyint=30:min-keyint=15:ref=1:bframes=2:b-pyramid=none\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"48\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1440,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 2560,\n"
"                    \"PresetDescription\": \"High quality video for publishing via online services such as Vimeo and YouTube. H.264 video (up to 1440p60) and high bit rate AAC stereo audio in an MP4 container.\",\n"
"                    \"PresetName\": \"Creator 1440p60 2.5K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 12000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"5.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"keyint=30:min-keyint=15:ref=1:bframes=2:b-pyramid=none\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 21.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"48\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"High quality video for publishing via online services such as Vimeo and YouTube. H.264 video (up to 1080p60) and high bit rate AAC stereo audio in an MP4 container.\",\n"
"                    \"PresetName\": \"Creator 1080p60\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 9000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"keyint=30:min-keyint=15:ref=1:bframes=2:b-pyramid=none\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"48\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"High quality video for publishing via online services such as Vimeo and YouTube. H.264 video (up to 720p60) and high bit rate AAC stereo audio in an MP4 container.\",\n"
"                    \"PresetName\": \"Creator 720p60\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 4500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"keyint=30:min-keyint=15:ref=1:bframes=2:b-pyramid=none\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 19.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"MetadataPassthru\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"Up to 30 seconds of video in 25 MB or less, for sharing via online social communities such as Discord and email services such as Gmail. H.264 video (up to 1080p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Social 25 MB 30 Seconds 1080p60\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"qcomp=0.5:aq-mode=2:aq-strength=0.8:bframes=2:b-adapt=2:b-pyramid=0\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 27.0,\n"
"                    \"VideoQualityType\": 1,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"MetadataPassthru\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"Up to 1 minute of video in 25 MB or less, for sharing via online social communities such as Discord and email services such as Gmail. H.264 video (up to 720p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Social 25 MB 1 Minute 720p60\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 3000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"qcomp=0.5:aq-mode=2:aq-strength=0.8:bframes=2:b-adapt=2:b-pyramid=0\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 27.0,\n"
"                    \"VideoQualityType\": 1,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"MetadataPassthru\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 540,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 960,\n"
"                    \"PresetDescription\": \"Up to 2 minutes of video in 25 MB or less, for sharing via online social communities such as Discord and email services such as Gmail. H.264 video (up to 540p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Social 25 MB 2 Minutes 540p60\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"qcomp=0.5:aq-mode=2:aq-strength=0.8:bframes=2:b-adapt=2:b-pyramid=0\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 27.0,\n"
"                    \"VideoQualityType\": 1,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"MetadataPassthru\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 360,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 640,\n"
"                    \"PresetDescription\": \"Up to 5 minutes of video in 25 MB or less, for sharing via online social communities such as Discord and email services such as Gmail. H.264 video (up to 360p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Social 25 MB 5 Minutes 360p60\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"qcomp=0.5:aq-mode=2:aq-strength=0.8:bframes=2:b-adapt=2:b-pyramid=0\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 27.0,\n"
"                    \"VideoQualityType\": 1,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"MetadataPassthru\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"Up to 30 seconds of video in 10 MB or less, for sharing via online social communities such as Discord and email services such as Gmail. H.264 video (up to 720p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Social 10 MB 30 Seconds 720p60\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 2250,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"qcomp=0.5:aq-mode=2:aq-strength=0.8:bframes=2:b-adapt=2:b-pyramid=0\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 27.0,\n"
"                    \"VideoQualityType\": 1,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"MetadataPassthru\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 540,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 960,\n"
"                    \"PresetDescription\": \"Up to 1 minute of video in 10 MB or less, for sharing via online social communities such as Discord and email services such as Gmail. H.264 video (up to 540p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Social 10 MB 1 Minute 540p60\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1125,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"qcomp=0.5:aq-mode=2:aq-strength=0.8:bframes=2:b-adapt=2:b-pyramid=0\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 27.0,\n"
"                    \"VideoQualityType\": 1,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"MetadataPassthru\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 360,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 640,\n"
"                    \"PresetDescription\": \"Up to 2 minutes of video in 10 MB or less, for sharing via online social communities such as Discord and email services such as Gmail. H.264 video (up to 360p60) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"Social 10 MB 2 Minutes 360p60\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"qcomp=0.5:aq-mode=2:aq-strength=0.8:bframes=2:b-adapt=2:b-pyramid=0\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 27.0,\n"
"                    \"VideoQualityType\": 1,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": true,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ],\n"
"            \"Folder\": true,\n"
"            \"PresetName\": \"Web\",\n"
"            \"Type\": 0\n"
"        },\n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) audio, in an MP4 container. Compatible with Amazon Fire TV 3rd Generation and later; Fire TV Stick 4K; Fire TV Cube.\",\n"
"                    \"PresetName\": \"Amazon Fire 2160p60 4K HEVC Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 8000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 24.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) audio, in an MP4 container. Compatible with Amazon Fire TV 1st Generation and later; Fire TV Stick 1st Generation and later; Fire HD 10 7th Generation (2017); Fire HDX 4th Generation (2014).\",\n"
"                    \"PresetName\": \"Amazon Fire 1080p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container. Compatible with Amazon Fire HD 4th Generation (2014) and later; Kindle Fire HDX 3rd Generation (2013); Kindle Fire HD 2nd Generation (2012) and later.\",\n"
"                    \"PresetName\": \"Amazon Fire 720p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 3000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 21.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30) and AAC stereo audio, in an MP4 container. Compatible with Android devices.\",\n"
"                    \"PresetName\": \"Android 1080p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container. Compatible with Android devices.\",\n"
"                    \"PresetName\": \"Android 720p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 3000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 21.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 128,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 576,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"H.264 video (up to 576p25) and AAC stereo audio, in an MP4 container. Compatible with Android devices.\",\n"
"                    \"PresetName\": \"Android 576p25\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1800,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"25\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 128,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 480,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"H.264 video (up to 480p30) and AAC stereo audio, in an MP4 container. Compatible with Android devices.\",\n"
"                    \"PresetName\": \"Android 480p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Apple iPhone 7 and later; Apple TV 4K.\",\n"
"                    \"PresetName\": \"Apple 2160p60 4K HEVC Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 8000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 24.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Apple iPad 5th and 6th Generation; iPad mini 2, 3, and 4; iPad Air 1st Generation and Air 2; iPad Pro 1st, 2nd, and 3rd Generation; Apple TV 4th Generation and later.\",\n"
"                    \"PresetName\": \"Apple 1080p60 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Apple iPhone 5, 5s, SE, 6, 6 Plus, 6s, 6s Plus, and later; iPod touch 6th Generation; iPad 3rd, 4th Generation and later; iPad mini 1st Generation and later; Apple TV 3rd, 4th Generation and later.\",\n"
"                    \"PresetName\": \"Apple 1080p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Apple iPhone 4, 4S, and later; iPod touch 4th, 5th Generation and later; iPad 1st Generation, iPad 2, and later; Apple TV 2nd Generation and later.\",\n"
"                    \"PresetName\": \"Apple 720p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 3000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 21.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 540,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 960,\n"
"                    \"PresetDescription\": \"H.264 video (up to 540p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Apple iPhone 1st Generation, 3G, 3GS, and later; iPod touch 1st, 2nd, 3rd Generation and later; iPod Classic; Apple TV 1st Generation and later.\",\n"
"                    \"PresetName\": \"Apple 540p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 2500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"qpmin=4:cabac=0:ref=2:b-pyramid=none:weightb=0:weightp=0:vbv-maxrate=9500:vbv-bufsize=9500\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Google Chromecast Ultra.\",\n"
"                    \"PresetName\": \"Chromecast 2160p60 4K HEVC Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 8000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 24.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Google Chromecast 3rd Generation.\",\n"
"                    \"PresetName\": \"Chromecast 1080p60 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": true,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Google Chromecast 1st, 2nd Generation and later.\",\n"
"                    \"PresetName\": \"Chromecast 1080p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"H.264 video (up to 2160p60), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Playstation 4 Pro.\",\n"
"                    \"PresetName\": \"Playstation 2160p60 4K Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 12000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"5.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 24.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Playstation 3 and 4.\",\n"
"                    \"PresetName\": \"Playstation 1080p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and AAC stereo audio, in an MP4 container. Compatible with Playstation Vita TV.\",\n"
"                    \"PresetName\": \"Playstation 720p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 3000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 21.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 540,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 960,\n"
"                    \"PresetDescription\": \"H.264 video (up to 540p30) and AAC stereo audio, in an MP4 container. Compatible with Playstation Vita.\",\n"
"                    \"PresetName\": \"Playstation 540p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 2250,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\",\n"
"                        \"copy:dts\",\n"
"                        \"copy:eac3\",\n"
"                        \"copy:mp3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60), AAC stereo audio, and surround audio, in an MKV container. Compatible with Roku 4, Streaming Stick+, Premiere+, and Ultra.\",\n"
"                    \"PresetName\": \"Roku 2160p60 4K HEVC Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 8000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 24.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Roku 1080p models.\",\n"
"                    \"PresetName\": \"Roku 1080p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Roku 720p models.\",\n"
"                    \"PresetName\": \"Roku 720p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 3000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 21.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 576,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"H.264 video (up to 576p25) and AAC stereo audio, in an MP4 container. Compatible with Roku standard definition models.\",\n"
"                    \"PresetName\": \"Roku 576p25\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1800,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"25\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": true,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 480,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"H.264 video (up to 480p30) and AAC stereo audio, in an MP4 container. Compatible with Roku standard definition models.\",\n"
"                    \"PresetName\": \"Roku 480p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        },\n"
"                        {\n"
"                            \"AudioBitrate\": 640,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy:ac3\",\n"
"                            \"AudioMixdown\": \"none\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30), AAC stereo audio, and Dolby Digital (AC-3) surround audio, in an MP4 container. Compatible with Xbox One.\",\n"
"                    \"PresetName\": \"Xbox 1080p30 Surround\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ],\n"
"            \"Folder\": true,\n"
"            \"PresetName\": \"Devices\",\n"
"            \"Type\": 0\n"
"        },\n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [],\n"
"                    \"AudioEncoderFallback\": \"opus\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"opus\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"AV1 video (up to 2160p60) and Opus stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"AV1 MKV 2160p60 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 10500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"svt_av1_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"enable-variance-boost=1:enable-qm=1:ac-bias=1:tf-strength=1:qp-scale-compress-strength=1:sharpness=1:keyint=10s\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"4\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 35.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"psnr\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"H.265 video (up to 2160p60) and AAC stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"H.265 MKV 2160p60 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 13000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 24.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.265 video (up to 1080p30) and AAC stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"H.265 MKV 1080p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 2500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"H.265 video (up to 720p30) and AAC stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"H.265 MKV 720p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 2000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 21.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 576,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"H.265 video (up to 576p25) and AAC stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"H.265 MKV 576p25\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1200,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"25\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 480,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"H.265 video (up to 480p30) and AAC stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"H.265 MKV 480p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x265_10bit\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"strong-intra-smoothing=0:rect=0:aq-mode=1:rd=4:psy-rd=0.75:psy-rdoq=4.0:rdoq-level=1:rskip=2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"slow\",\n"
"                    \"VideoProfile\": \"main10\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"H.264 video (up to 2160p60) and AAC stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"H.264 MKV 2160p60 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 12000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"5.2\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 24.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"H.264 video (up to 1080p30) and AAC stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"H.264 MKV 1080p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"4.0\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"H.264 video (up to 720p30) and AAC stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"H.264 MKV 720p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 3000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 21.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 576,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"H.264 video (up to 576p25) and AAC stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"H.264 MKV 576p25\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1800,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"25\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 480,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"H.264 video (up to 480p30) and AAC stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"H.264 MKV 480p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"3.1\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 20.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [],\n"
"                    \"AudioEncoderFallback\": \"opus\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"opus\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"VP9 video (up to 2160p60) and Opus stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"VP9 MKV 2160p60 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 8000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"vp9\",\n"
"                    \"VideoFramerate\": \"60\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 30.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [],\n"
"                    \"AudioEncoderFallback\": \"opus\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"opus\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"VP9 video (up to 1080p30) and Opus stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"VP9 MKV 1080p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 4500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"vp9\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 30.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [],\n"
"                    \"AudioEncoderFallback\": \"opus\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"opus\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 720,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1280,\n"
"                    \"PresetDescription\": \"VP9 video (up to 720p30) and Opus stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"VP9 MKV 720p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 2500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"vp9\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 30.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [],\n"
"                    \"AudioEncoderFallback\": \"opus\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"opus\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 576,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"VP9 video (up to 576p25) and Opus stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"VP9 MKV 576p25\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"vp9\",\n"
"                    \"VideoFramerate\": \"25\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 30.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [],\n"
"                    \"AudioEncoderFallback\": \"opus\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"opus\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"default\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"decomb\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 480,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 720,\n"
"                    \"PresetDescription\": \"VP9 video (up to 480p30) and Opus stereo audio, in an MKV container.\",\n"
"                    \"PresetName\": \"VP9 MKV 480p30\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": true,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 1250,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"vp9\",\n"
"                    \"VideoFramerate\": \"30\",\n"
"                    \"VideoFramerateMode\": \"pfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 30.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ],\n"
"            \"Folder\": true,\n"
"            \"PresetName\": \"Matroska\",\n"
"            \"Type\": 0\n"
"        },\n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"Intel Quick Sync Video hardware accelerated AV1 video (up to 2160p) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"AV1 QSV 2160p 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 6500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"qsv_av1_10bit\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"speed\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 24.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"AMD VCN hardware accelerated AV1 video (up to 2160p) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"AV1 VCN 2160p 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 10000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"vce_av1_10bit\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"balanced\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 120.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"Nvidia NVENC hardware accelerated H.265 video (up to 2160p) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"H.265 NVENC 2160p 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 10500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"nvenc_h265\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"rc-lookahead=10\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 30.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"Nvidia NVENC hardware accelerated H.265 video (up to 1080p) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"H.265 NVENC 1080p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 3750,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"nvenc_h265\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"rc-lookahead=10\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 27.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"Intel Quick Sync Video hardware accelerated H.265 video (up to 2160p) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"H.265 QSV 2160p 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 8000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"qsv_h265\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"speed\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 28.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"Intel Quick Sync Video hardware accelerated H.265 video (up to 1080p) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"H.265 QSV 1080p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 2500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"qsv_h265\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"speed\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 25.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"AMD VCN hardware accelerated H.265 video (up to 2160p) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"H.265 VCN 2160p 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 12500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"vce_h265\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"balanced\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 28.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"AMD VCN hardware accelerated H.265 video (up to 1080p) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"H.265 VCN 1080p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 7250,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"vce_h265\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"balanced\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 25.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 1,\n"
"                    \"PicturePARWidth\": 1,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"Hardware accelerated H.265 video (up to 2160p) and AAC stereo audio, in an MP4 container for ARM based platforms using Media Foundation\",\n"
"                    \"PresetName\": \"H.265 MF 2160p 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 9500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"mf_h265\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"balanced\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 28.0,\n"
"                    \"VideoQualityType\": 1,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 1,\n"
"                    \"PicturePARWidth\": 1,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"Hardware accelerated H.265 video (up to 1080p) and AAC stereo audio, in an MP4 container for ARM based platforms using Media Foundation\",\n"
"                    \"PresetName\": \"H.265 MF 1080p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 5500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"mf_h265\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"balanced\",\n"
"                    \"VideoProfile\": \"main\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 1,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureLooseCrop\": false,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"Apple VideoToolbox hardware accelerated H.265 video (up to 2160p) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"H.265 Apple VideoToolbox 2160p 4K\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 17500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"vt_h265_10bit\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"speed\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 55.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 160,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureLooseCrop\": false,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"Apple VideoToolbox hardware accelerated H.265 video (up to 1080p) and AAC stereo audio, in an MP4 container.\",\n"
"                    \"PresetName\": \"H.265 Apple VideoToolbox 1080p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 5500,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"vt_h265_10bit\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"speed\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 60.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ],\n"
"            \"Folder\": true,\n"
"            \"PresetName\": \"Hardware\",\n"
"            \"Type\": 0\n"
"        },\n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"Maximum bit rate, constant frame rate H.264 video and high bit rate AAC stereo audio in an MP4 container. For professional use as an intermediate format for video editing. Creates gigantic files.\",\n"
"                    \"PresetName\": \"Production Max\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 120000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"auto\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"keyint=12:min-keyint=1:ref=1:bframes=0:qcomp=0.8:aq-strength=0.5:dct-decimate=0:fast-pskip=0:deblock=-2,-2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"fast\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 2.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"High bit rate, constant frame rate H.264 video and high bit rate AAC stereo audio in an MP4 container. For professional use as an intermediate format for video editing. Creates very large files.\",\n"
"                    \"PresetName\": \"Production Standard\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 60000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"auto\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"keyint=12:min-keyint=1:ref=1:bframes=0:qcomp=0.8:aq-strength=0.5:dct-decimate=0:fast-pskip=0:deblock=-2,-2\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"fast\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 10.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"Intra-only, constant frame rate H.264 video (up to 1080p) and high bit rate AAC stereo audio in an MP4 container. For professional use as a low resolution proxy format for video editing.\",\n"
"                    \"PresetName\": \"Production Proxy 1080p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 30000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"auto\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"keyint=1:min-keyint=1:ref=1:bframes=0:qcomp=0.8:aq-strength=0.5\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"superfast\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 18.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"none\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 320,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"stereo\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 540,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 960,\n"
"                    \"PresetDescription\": \"Intra-only, constant frame rate H.264 video (up to 540p) and high bit rate AAC stereo audio in an MP4 container. For professional use as a low resolution proxy format for video editing.\",\n"
"                    \"PresetName\": \"Production Proxy 540p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 15000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"auto\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"keyint=1:min-keyint=1:ref=1:bframes=0:qcomp=0.8:aq-strength=0.5\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"superfast\",\n"
"                    \"VideoProfile\": \"high\",\n"
"                    \"VideoQualitySlider\": 16.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"Maximum bit rate, constant frame rate DNxHR HQX video and PCM audio in a MOV container. For professional use as an intermediate format for video editing. Creates gigantic files.\",\n"
"                    \"PresetName\": \"Production DNxHR HQX\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 220000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"dnxhr_10bit\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"hqx\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"Very high bit rate, constant frame rate DNxHR SQ video and PCM audio in a MOV container. For professional use as an intermediate format for video editing. Creates gigantic files.\",\n"
"                    \"PresetName\": \"Production DNxHR SQ\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 145000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"dnxhr\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"sq\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"Constant frame rate DNxHR LB video (up to 2160p) and PCM audio in a MOV container. For professional use as a low resolution proxy format for video editing. Creates very large files.\",\n"
"                    \"PresetName\": \"Production DNxHR Proxy 2160p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 179000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"dnxhr\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"lb\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"Constant frame rate DNxHR LB video (up to 1080p) and PCM audio in a MOV container. For professional use as a low resolution proxy format for video editing. Creates very large files.\",\n"
"                    \"PresetName\": \"Production DNxHR Proxy 1080p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 45000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"dnxhr\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"lb\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 540,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 960,\n"
"                    \"PresetDescription\": \"Constant frame rate DNxHR LB video (up to 540p) and PCM audio in a MOV container. For professional use as a low resolution proxy format for video editing.\",\n"
"                    \"PresetName\": \"Production DNxHR Proxy 540p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 11000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"dnxhr\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"lb\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"Maximum bit rate, constant frame rate ProRes 422 HQ video and PCM audio in a MOV container. For professional use as an intermediate format for video editing. Creates gigantic files.\",\n"
"                    \"PresetName\": \"Production ProRes 422 HQ\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 220000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"prores\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"hq\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"Very high bit rate, constant frame rate ProRes 422 video and PCM audio in a MOV container. For professional use as an intermediate format for video editing. Creates gigantic files.\",\n"
"                    \"PresetName\": \"Production ProRes 422\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 147000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"prores\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"standard\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"High bit rate, constant frame rate ProRes 422 LT video and PCM audio in a MOV container. For professional use as an intermediate format for video editing. Creates very large files.\",\n"
"                    \"PresetName\": \"Production ProRes 422 LT\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 102000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"prores\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"lt\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 2160,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 3840,\n"
"                    \"PresetDescription\": \"Constant frame rate ProRes Proxy video (up to 2160p) and PCM audio in a MOV container. For professional use as a low resolution proxy format for video editing. Creates very large files.\",\n"
"                    \"PresetName\": \"Production ProRes Proxy 2160p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 160000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"prores\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"proxy\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 1080,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 1920,\n"
"                    \"PresetDescription\": \"Constant frame rate ProRes Proxy video (up to 1080p) and PCM audio in a MOV container. For professional use as a low resolution proxy format for video editing. Creates very large files.\",\n"
"                    \"PresetName\": \"Production ProRes Proxy 1080p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 45000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"prores\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"proxy\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"pcm24\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mov\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 540,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 960,\n"
"                    \"PresetDescription\": \"Constant frame rate ProRes Proxy video (up to 540p) and PCM audio in a MOV container. For professional use as a low resolution proxy format for video editing.\",\n"
"                    \"PresetName\": \"Production ProRes Proxy 540p\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": true,\n"
"                    \"SubtitleBurnBehavior\": \"foreign\",\n"
"                    \"SubtitleBurnDVDSub\": true,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 12000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"prores\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"cfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"auto\",\n"
"                    \"VideoProfile\": \"proxy\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\",\n"
"                        \"copy:eac3\",\n"
"                        \"copy:truehd\",\n"
"                        \"copy:dts\",\n"
"                        \"copy:dtshd\",\n"
"                        \"copy:mp2\",\n"
"                        \"copy:mp3\",\n"
"                        \"copy:flac\",\n"
"                        \"copy:opus\",\n"
"                        \"copy:vorbis\",\n"
"                        \"copy:alac\",\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"flac24\",\n"
"                    \"AudioLanguageList\": [\n"
"                        \"any\"\n"
"                    ],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"copy\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"Lossless FFV1 video and original audio with FLAC fallback in an MKV container. For professional use as an archival format for digital preservation. Creates extremely large files.\",\n"
"                    \"PresetName\": \"Preservation FFV1\",\n"
"                    \"SubtitleAddCC\": true,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [\n"
"                        \"any\"\n"
"                    ],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"all\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 0,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"auto\",\n"
"                    \"VideoEncoder\": \"ffv1\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"preservation\",\n"
"                    \"VideoProfile\": \"\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:flac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"flac24\",\n"
"                    \"AudioLanguageList\": [\n"
"                        \"any\"\n"
"                    ],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"flac24\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"Lossless FFV1 video and FLAC audio in an MKV container. For professional use as an archival format for digital preservation. Creates extremely large files.\",\n"
"                    \"PresetName\": \"Preservation FFV1 FLAC\",\n"
"                    \"SubtitleAddCC\": true,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [\n"
"                        \"any\"\n"
"                    ],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"all\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 0,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"auto\",\n"
"                    \"VideoEncoder\": \"ffv1\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"preservation\",\n"
"                    \"VideoProfile\": \"\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                },\n"
"                {\n"
"                    \"AlignAVStart\": false,\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:pcm\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"flac24\",\n"
"                    \"AudioLanguageList\": [\n"
"                        \"any\"\n"
"                    ],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": -1.0,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"pcm24\",\n"
"                            \"AudioMixdown\": \"7point1\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": false,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"all\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": false,\n"
"                    \"FileFormat\": \"mkv\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"InlineParameterSets\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 2,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"light\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"off\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"Lossless FFV1 video and PCM audio in an MKV container. For professional use as an archival format for digital preservation. Creates extremely large files.\",\n"
"                    \"PresetName\": \"Preservation FFV1 PCM\",\n"
"                    \"SubtitleAddCC\": true,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [\n"
"                        \"any\"\n"
"                    ],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"all\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 1,\n"
"                    \"VideoAvgBitrate\": 0,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"auto\",\n"
"                    \"VideoEncoder\": \"ffv1\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": true,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"preservation\",\n"
"                    \"VideoProfile\": \"\",\n"
"                    \"VideoQualitySlider\": 0.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ],\n"
"            \"Folder\": true,\n"
"            \"PresetName\": \"Professional\",\n"
"            \"Type\": 0\n"
"        }\n"
"    ],\n"
"    \"PresetCLIDefault\": [\n"
"        {\n"
"            \"ChildrenArray\": [\n"
"                {\n"
"                    \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"                    \"AudioCopyMask\": [\n"
"                        \"copy:aac\",\n"
"                        \"copy:ac3\",\n"
"                        \"copy:eac3\",\n"
"                        \"copy:dtshd\",\n"
"                        \"copy:dts\",\n"
"                        \"copy:mp3\",\n"
"                        \"copy:truehd\",\n"
"                        \"copy:flac\"\n"
"                    ],\n"
"                    \"AudioEncoderFallback\": \"aac\",\n"
"                    \"AudioLanguageList\": [],\n"
"                    \"AudioList\": [\n"
"                        {\n"
"                            \"AudioBitrate\": 128,\n"
"                            \"AudioCompressionLevel\": -1.0,\n"
"                            \"AudioDitherMethod\": \"auto\",\n"
"                            \"AudioEncoder\": \"aac\",\n"
"                            \"AudioMixdown\": \"dpl2\",\n"
"                            \"AudioNormalizeMixLevel\": false,\n"
"                            \"AudioSamplerate\": \"auto\",\n"
"                            \"AudioTrackDRCSlider\": 0.0,\n"
"                            \"AudioTrackGainSlider\": 0.0,\n"
"                            \"AudioTrackQuality\": -1.0,\n"
"                            \"AudioTrackQualityEnable\": false\n"
"                        }\n"
"                    ],\n"
"                    \"AudioSecondaryEncoderMode\": true,\n"
"                    \"AudioTrackNamePassthru\": true,\n"
"                    \"AudioTrackSelectionBehavior\": \"first\",\n"
"                    \"ChapterMarkers\": true,\n"
"                    \"ChildrenArray\": [],\n"
"                    \"Default\": true,\n"
"                    \"FileFormat\": \"mp4\",\n"
"                    \"Folder\": false,\n"
"                    \"FolderOpen\": false,\n"
"                    \"Mp4iPodCompatible\": false,\n"
"                    \"Optimize\": false,\n"
"                    \"PictureBottomCrop\": 0,\n"
"                    \"PictureChromaSmoothCustom\": \"\",\n"
"                    \"PictureChromaSmoothPreset\": \"off\",\n"
"                    \"PictureChromaSmoothTune\": \"none\",\n"
"                    \"PictureColorspaceCustom\": \"\",\n"
"                    \"PictureColorspacePreset\": \"off\",\n"
"                    \"PictureCombDetectCustom\": \"\",\n"
"                    \"PictureCombDetectPreset\": \"off\",\n"
"                    \"PictureCropMode\": 0,\n"
"                    \"PictureDARWidth\": 0,\n"
"                    \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"                    \"PictureDeblockPreset\": \"off\",\n"
"                    \"PictureDeblockTune\": \"medium\",\n"
"                    \"PictureDeinterlaceCustom\": \"\",\n"
"                    \"PictureDeinterlaceFilter\": \"off\",\n"
"                    \"PictureDeinterlacePreset\": \"default\",\n"
"                    \"PictureDenoiseCustom\": \"\",\n"
"                    \"PictureDenoiseFilter\": \"off\",\n"
"                    \"PictureDenoisePreset\": \"medium\",\n"
"                    \"PictureDenoiseTune\": \"none\",\n"
"                    \"PictureDetelecine\": \"off\",\n"
"                    \"PictureDetelecineCustom\": \"\",\n"
"                    \"PictureForceHeight\": 0,\n"
"                    \"PictureForceWidth\": 0,\n"
"                    \"PictureHeight\": 0,\n"
"                    \"PictureItuPAR\": false,\n"
"                    \"PictureKeepRatio\": true,\n"
"                    \"PictureLeftCrop\": 0,\n"
"                    \"PictureModulus\": 2,\n"
"                    \"PicturePAR\": \"auto\",\n"
"                    \"PicturePARHeight\": 720,\n"
"                    \"PicturePARWidth\": 853,\n"
"                    \"PictureRightCrop\": 0,\n"
"                    \"PictureRotate\": \"angle=0:hflip=0\",\n"
"                    \"PictureSharpenCustom\": \"\",\n"
"                    \"PictureSharpenFilter\": \"off\",\n"
"                    \"PictureSharpenPreset\": \"medium\",\n"
"                    \"PictureSharpenTune\": \"none\",\n"
"                    \"PictureTopCrop\": 0,\n"
"                    \"PictureWidth\": 0,\n"
"                    \"PresetDescription\": \"\",\n"
"                    \"PresetName\": \"CLI Default\",\n"
"                    \"SubtitleAddCC\": false,\n"
"                    \"SubtitleAddForeignAudioSearch\": false,\n"
"                    \"SubtitleAddForeignAudioSubtitle\": false,\n"
"                    \"SubtitleBurnBDSub\": false,\n"
"                    \"SubtitleBurnBehavior\": \"none\",\n"
"                    \"SubtitleBurnDVDSub\": false,\n"
"                    \"SubtitleLanguageList\": [],\n"
"                    \"SubtitleTrackNamePassthru\": true,\n"
"                    \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"                    \"Type\": 0,\n"
"                    \"UsesPictureFilters\": true,\n"
"                    \"UsesPictureSettings\": 0,\n"
"                    \"VideoAvgBitrate\": 6000,\n"
"                    \"VideoColorMatrixCodeOverride\": 0,\n"
"                    \"VideoColorRange\": \"limited\",\n"
"                    \"VideoEncoder\": \"x264\",\n"
"                    \"VideoFramerate\": \"auto\",\n"
"                    \"VideoFramerateMode\": \"vfr\",\n"
"                    \"VideoGrayScale\": false,\n"
"                    \"VideoHWDecode\": 0,\n"
"                    \"VideoLevel\": \"auto\",\n"
"                    \"VideoMultiPass\": false,\n"
"                    \"VideoOptionExtra\": \"\",\n"
"                    \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"                    \"VideoPreset\": \"medium\",\n"
"                    \"VideoProfile\": \"auto\",\n"
"                    \"VideoQualitySlider\": 22.0,\n"
"                    \"VideoQualityType\": 2,\n"
"                    \"VideoScaler\": \"swscale\",\n"
"                    \"VideoTune\": \"\",\n"
"                    \"VideoTurboMultiPass\": false,\n"
"                    \"x264Option\": \"\",\n"
"                    \"x264UseAdvancedOptions\": false\n"
"                }\n"
"            ],\n"
"            \"Folder\": true,\n"
"            \"PresetName\": \"CLI Defaults\",\n"
"            \"Type\": 0\n"
"        }\n"
"    ],\n"
"    \"PresetTemplate\": {\n"
"        \"Preset\": {\n"
"            \"AlignAVStart\": false,\n"
"            \"AudioAutomaticNamingBehavior\": \"unnamed\",\n"
"            \"AudioCopyMask\": [\n"
"                \"copy:aac\",\n"
"                \"copy:ac3\",\n"
"                \"copy:dts\",\n"
"                \"copy:dtshd\",\n"
"                \"copy:eac3\",\n"
"                \"copy:flac\",\n"
"                \"copy:mp3\",\n"
"                \"copy:truehd\"\n"
"            ],\n"
"            \"AudioEncoderFallback\": \"ac3\",\n"
"            \"AudioLanguageList\": [\n"
"                \"any\"\n"
"            ],\n"
"            \"AudioList\": [\n"
"                {\n"
"                    \"AudioBitrate\": 192,\n"
"                    \"AudioCompressionLevel\": -1.0,\n"
"                    \"AudioDitherMethod\": \"auto\",\n"
"                    \"AudioEncoder\": \"copy:ac3\",\n"
"                    \"AudioMixdown\": \"dpl2\",\n"
"                    \"AudioNormalizeMixLevel\": false,\n"
"                    \"AudioSamplerate\": \"auto\",\n"
"                    \"AudioTrackDRCSlider\": 0.0,\n"
"                    \"AudioTrackGainSlider\": 0.0,\n"
"                    \"AudioTrackQuality\": -1.0,\n"
"                    \"AudioTrackQualityEnable\": false\n"
"                }\n"
"            ],\n"
"            \"AudioSecondaryEncoderMode\": true,\n"
"            \"AudioTrackNamePassthru\": true,\n"
"            \"AudioTrackSelectionBehavior\": \"first\",\n"
"            \"ChapterMarkers\": true,\n"
"            \"ChildrenArray\": [],\n"
"            \"Default\": false,\n"
"            \"FileFormat\": \"mp4\",\n"
"            \"Folder\": false,\n"
"            \"FolderOpen\": false,\n"
"            \"InlineParameterSets\": false,\n"
"            \"MetadataPassthru\": true,\n"
"            \"Mp4iPodCompatible\": false,\n"
"            \"Optimize\": false,\n"
"            \"PictureAllowUpscaling\": false,\n"
"            \"PictureAutoCrop\": true,\n"
"            \"PictureBottomCrop\": 0,\n"
"            \"PictureChromaSmoothCustom\": \"\",\n"
"            \"PictureChromaSmoothPreset\": \"off\",\n"
"            \"PictureChromaSmoothTune\": \"none\",\n"
"            \"PictureColorspaceCustom\": \"\",\n"
"            \"PictureColorspacePreset\": \"off\",\n"
"            \"PictureCombDetectCustom\": \"\",\n"
"            \"PictureCombDetectPreset\": \"off\",\n"
"            \"PictureCropMode\": 0,\n"
"            \"PictureDARWidth\": 0,\n"
"            \"PictureDeblockCustom\": \"strength=strong:thresh=20:blocksize=8\",\n"
"            \"PictureDeblockPreset\": \"off\",\n"
"            \"PictureDeblockTune\": \"medium\",\n"
"            \"PictureDeinterlaceCustom\": \"\",\n"
"            \"PictureDeinterlaceFilter\": \"off\",\n"
"            \"PictureDeinterlacePreset\": \"default\",\n"
"            \"PictureDenoiseCustom\": \"\",\n"
"            \"PictureDenoiseFilter\": \"off\",\n"
"            \"PictureDenoisePreset\": \"medium\",\n"
"            \"PictureDenoiseTune\": \"none\",\n"
"            \"PictureDetelecine\": \"off\",\n"
"            \"PictureDetelecineCustom\": \"\",\n"
"            \"PictureForceHeight\": 0,\n"
"            \"PictureForceWidth\": 0,\n"
"            \"PictureHeight\": 0,\n"
"            \"PictureItuPAR\": false,\n"
"            \"PictureKeepRatio\": true,\n"
"            \"PictureLeftCrop\": 0,\n"
"            \"PictureModulus\": 2,\n"
"            \"PicturePAR\": \"auto\",\n"
"            \"PicturePARHeight\": 720,\n"
"            \"PicturePARWidth\": 853,\n"
"            \"PicturePadBottom\": 0,\n"
"            \"PicturePadColor\": \"black\",\n"
"            \"PicturePadLeft\": 0,\n"
"            \"PicturePadMode\": \"none\",\n"
"            \"PicturePadRight\": 0,\n"
"            \"PicturePadTop\": 0,\n"
"            \"PictureRightCrop\": 0,\n"
"            \"PictureRotate\": \"angle=0:hflip=0\",\n"
"            \"PictureSharpenCustom\": \"\",\n"
"            \"PictureSharpenFilter\": \"off\",\n"
"            \"PictureSharpenPreset\": \"medium\",\n"
"            \"PictureSharpenTune\": \"none\",\n"
"            \"PictureTopCrop\": 0,\n"
"            \"PictureUseMaximumSize\": true,\n"
"            \"PictureWidth\": 0,\n"
"            \"PresetDescription\": \"\",\n"
"            \"PresetDisabled\": false,\n"
"            \"PresetName\": \"Name Missing\",\n"
"            \"SubtitleAddCC\": false,\n"
"            \"SubtitleAddForeignAudioSearch\": false,\n"
"            \"SubtitleAddForeignAudioSubtitle\": false,\n"
"            \"SubtitleBurnBDSub\": false,\n"
"            \"SubtitleBurnBehavior\": \"none\",\n"
"            \"SubtitleBurnDVDSub\": false,\n"
"            \"SubtitleLanguageList\": [],\n"
"            \"SubtitleTrackNamePassthru\": true,\n"
"            \"SubtitleTrackSelectionBehavior\": \"none\",\n"
"            \"Type\": 1,\n"
"            \"UsesPictureFilters\": true,\n"
"            \"VideoAvgBitrate\": 1800,\n"
"            \"VideoColorMatrixCodeOverride\": 0,\n"
"            \"VideoColorRange\": \"limited\",\n"
"            \"VideoEncoder\": \"x264\",\n"
"            \"VideoFramerate\": \"auto\",\n"
"            \"VideoFramerateMode\": \"vfr\",\n"
"            \"VideoGrayScale\": false,\n"
"            \"VideoHWDecode\": 0,\n"
"            \"VideoLevel\": \"auto\",\n"
"            \"VideoMultiPass\": false,\n"
"            \"VideoOptionExtra\": \"\",\n"
"            \"VideoPasshtruHDRDynamicMetadata\": \"all\",\n"
"            \"VideoPreset\": \"medium\",\n"
"            \"VideoProfile\": \"auto\",\n"
"            \"VideoQualitySlider\": 20.0,\n"
"            \"VideoQualityType\": 2,\n"
"            \"VideoScaler\": \"swscale\",\n"
"            \"VideoTune\": \"\",\n"
"            \"VideoTurboMultiPass\": false,\n"
"            \"x264Option\": \"\",\n"
"            \"x264UseAdvancedOptions\": false\n"
"        },\n"
"        \"VersionMajor\": 72,\n"
"        \"VersionMicro\": 0,\n"
"        \"VersionMinor\": 0\n"
"    }\n"
"}\n";
